{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Question 78 - LC 173. Binary Search Tree Iterator\n",
    "\n",
    "Medium\n",
    "\n",
    "## Topics\n",
    "\n",
    "- Design\n",
    "- Stack\n",
    "- Tree\n",
    "- Binary Search Tree\n",
    "- Binary Tree\n",
    "- Iterator\n",
    "\n",
    "## Companies\n",
    "\n",
    "- Amazon\n",
    "- Facebook\n",
    "- Google\n",
    "- Microsoft\n",
    "\n",
    "## Description\n",
    "\n",
    "Implement the `BSTIterator` class that represents an iterator over the in-order traversal of a binary search tree (BST):\n",
    "\n",
    "- `BSTIterator(TreeNode root)` Initializes an object of the `BSTIterator` class. The root of the BST is given as part of the constructor. The pointer should be initialized to a non-existent number smaller than any element in the BST.\n",
    "- `boolean hasNext()` Returns `true` if there exists a number in the traversal to the right of the pointer, otherwise returns `false`.\n",
    "- `int next()` Moves the pointer to the right, then returns the number at the pointer.\n",
    "\n",
    "Notice that by initializing the pointer to a non-existent smallest number, the first call to `next()` will return the smallest element in the BST.\n",
    "\n",
    "You may assume that `next()` calls will always be valid. That is, there will be at least a next number in the in-order traversal when `next()` is called.\n",
    "\n",
    "### Example 1:\n",
    "\n",
    "```\n",
    "Input\n",
    "[\"BSTIterator\", \"next\", \"next\", \"hasNext\", \"next\", \"hasNext\", \"next\", \"hasNext\", \"next\", \"hasNext\"]\n",
    "[[[7, 3, 15, null, null, 9, 20]], [], [], [], [], [], [], [], [], []]\n",
    "Output\n",
    "[null, 3, 7, true, 9, true, 15, true, 20, false]\n",
    "\n",
    "Explanation\n",
    "BSTIterator bSTIterator = new BSTIterator([7, 3, 15, null, null, 9, 20]);\n",
    "bSTIterator.next();    // return 3\n",
    "bSTIterator.next();    // return 7\n",
    "bSTIterator.hasNext(); // return True\n",
    "bSTIterator.next();    // return 9\n",
    "bSTIterator.hasNext(); // return True\n",
    "bSTIterator.next();    // return 15\n",
    "bSTIterator.hasNext(); // return True\n",
    "bSTIterator.next();    // return 20\n",
    "bSTIterator.hasNext(); // return False\n",
    "```\n",
    "\n",
    "### Constraints:\n",
    "\n",
    "- The number of nodes in the tree is in the range [1, 105].\n",
    "- 0 <= Node.val <= 106\n",
    "- At most 105 calls will be made to `hasNext`, and `next`.\n",
    "\n",
    "### Follow up:\n",
    "\n",
    "Could you implement `next()` and `hasNext()` to run in average O(1) time and use O(h) memory, where h is the height of the tree?\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "from __future__ import annotations\n",
    "from typing import Optional\n",
    "from collections import deque"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Definition for a binary tree node.\n",
    "class TreeNode:\n",
    "    def __init__(\n",
    "        self,\n",
    "        val: int = 0,\n",
    "        left: Optional[TreeNode] = None,\n",
    "        right: Optional[TreeNode] = None,\n",
    "    ):\n",
    "        self.val = val\n",
    "        self.left = left\n",
    "        self.right = right\n",
    "\n",
    "    @staticmethod\n",
    "    def from_list(values: list[Optional[int]]) -> Optional[TreeNode]:\n",
    "        \"\"\"\n",
    "        Constructs a binary tree from a list of values using level-order traversal (BFS).\n",
    "\n",
    "        Parameters:\n",
    "        - values (list): The list of values to construct the binary tree from.\n",
    "\n",
    "        Returns:\n",
    "        TreeNode: The root of the constructed binary tree.\n",
    "        \"\"\"\n",
    "        if not values:\n",
    "            return None\n",
    "\n",
    "        root = TreeNode(val=values[0])\n",
    "        queue = deque([root])\n",
    "        i = 1\n",
    "\n",
    "        while queue and i < len(values):\n",
    "            current_node = queue.popleft()\n",
    "\n",
    "            if values[i] is not None:\n",
    "                current_node.left = TreeNode(val=values[i])\n",
    "                queue.append(current_node.left)\n",
    "            i += 1\n",
    "\n",
    "            if i < len(values) and values[i] is not None:\n",
    "                current_node.right = TreeNode(val=values[i])\n",
    "                queue.append(current_node.right)\n",
    "            i += 1\n",
    "\n",
    "        return root\n",
    "\n",
    "    def print_level_order(self):\n",
    "        \"\"\"\n",
    "        Prints the level order traversal of the tree (BFS).\n",
    "\n",
    "        Returns:\n",
    "        list: A list of lists, where each sublist contains the values of the tree nodes\n",
    "              at that depth.\n",
    "        \"\"\"\n",
    "        queue = deque([(self, 0)])\n",
    "        result = []\n",
    "        current_level = []\n",
    "        level_number = 0\n",
    "\n",
    "        while queue:\n",
    "            current_node, node_level = queue.popleft()\n",
    "\n",
    "            if node_level > level_number:\n",
    "                result.append(current_level)\n",
    "                current_level = []\n",
    "                level_number = node_level\n",
    "\n",
    "            if current_node:\n",
    "                current_level.append(current_node.val)\n",
    "                queue.append((current_node.left, node_level + 1))\n",
    "                queue.append((current_node.right, node_level + 1))\n",
    "            else:\n",
    "                current_level.append(\"None\")\n",
    "\n",
    "        if current_level:\n",
    "            result.append(current_level)\n",
    "\n",
    "        # Remove trailing \"None\" values\n",
    "        result = [level for level in result if any(elem != \"None\" for elem in level)]\n",
    "\n",
    "        return result"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [],
   "source": [
    "class BSTIterator:\n",
    "    def __init__(self, root: Optional[TreeNode]):\n",
    "        self.stack: list[TreeNode] = []\n",
    "        self._leftmost_inorder(root)\n",
    "\n",
    "    def _leftmost_inorder(self, node: Optional[TreeNode]):\n",
    "        # Push all nodes from root to the left-most node\n",
    "        while node:\n",
    "            self.stack.append(node)\n",
    "            node = node.left\n",
    "\n",
    "    def next(self) -> int:\n",
    "        \"\"\"\n",
    "        @return the next smallest number\n",
    "        \"\"\"\n",
    "        # Pop the top node from the stack, process it, and then push all its right node's left children\n",
    "        topmost_node = self.stack.pop()\n",
    "        self._leftmost_inorder(topmost_node.right)\n",
    "\n",
    "        return topmost_node.val\n",
    "\n",
    "    def hasNext(self) -> bool:\n",
    "        \"\"\"\n",
    "        @return whether we have a next smallest number\n",
    "        \"\"\"\n",
    "        return len(self.stack) > 0"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "3\n",
      "7\n",
      "True\n",
      "9\n",
      "True\n",
      "15\n",
      "True\n",
      "20\n",
      "False\n"
     ]
    }
   ],
   "source": [
    "# Creating a BST from a list\n",
    "tree_list = [7, 3, 15, None, None, 9, 20]\n",
    "node = TreeNode.from_list(tree_list)\n",
    "\n",
    "# Initializing BSTIterator with the constructed BST\n",
    "iterator = BSTIterator(node)\n",
    "\n",
    "# Making calls to next() and hasNext() and printing the outputs\n",
    "print(iterator.next())    # Expected output: 3\n",
    "print(iterator.next())    # Expected output: 7\n",
    "print(iterator.hasNext())  # Expected output: True\n",
    "print(iterator.next())    # Expected output: 9\n",
    "print(iterator.hasNext())  # Expected output: True\n",
    "print(iterator.next())    # Expected output: 15\n",
    "print(iterator.hasNext())  # Expected output: True\n",
    "print(iterator.next())    # Expected output: 20\n",
    "print(iterator.hasNext())  # Expected output: False"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.12.0"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
